
from time import time
from urllib.parse import urlencode
from django.conf import settings
from django.core.signing import Signer
from django.core.urlresolvers import reverse
from django.middleware.csrf import get_token
from django.utils.safestring import mark_safe
from os.path import join, dirname
from re import sub


signer = Signer()


def minimal_javascript_log(request):
	"""
	Create the url that javascript logs can be sent to. There's no way to prevent the client from sending fake log
	files, since any credentials must be in the code. But the damage can be limited by requiring csrf/signed value
	(so only the actual client) and timestamp (so they can only send logs for a limited time), rather than giving
	them direct access to the logging server.
	"""
	url = reverse('minimal_log_receive_js_log')
	if settings.APPEND_SLASH and not url.endswith('/'):
		url += '/'
	params = urlencode({
		'csrfmiddlewaretoken': get_token(request),
		'request_time': signer.sign('{0:.0f}'.format(time()))
	})
	with open(join(dirname(__file__), 'js_handler.html')) as fh:
		code = fh.read().strip()
		code = sub(r'{{\s*MINIMAL_LOG_JS_URL\s*}}',    url,    code)
		code = sub(r'{{\s*MINIMAL_LOG_JS_PARAMS\s*}}', params, code)
	return {
		'MINIMAL_LOG_JS_URL': mark_safe(url),
		'MINIMAL_LOG_JS_PARAMS': mark_safe(params),
		'MINIMAL_LOG_JAVASCRIPT_HANDLING_CODE': mark_safe(code),
	}






from django.conf import settings
from os.path import basename
from requests import post


class MinimalLogError(Exception):
	pass


def try_get_domain():
	from django.contrib.sites.models import Site
	try:
		domain = Site.objects.get_current().domain
	except:
		pass
	else:
		if not 'example' in domain:
			return domain
	dr = getattr(settings, 'BASE_DIR', None)
	if dr:
		return basename(dr)
	return 'unknown-django-project'


def send_log(message, key=None, url=None, status='error', check_status=True):
	if key is None:
		assert hasattr(settings, 'MINIMAL_LOG_AUTH_KEY'), \
			'send_log did not get a key and could not find one in Django settings (MINIMAL_LOG_AUTH_KEY)'
		key = settings.MINIMAL_LOG_AUTH_KEY
	if url is None:
		assert hasattr(settings, 'MINIMAL_LOG_URL'), \
			'send_log did not get a url and could not find one in Django settings (MINIMAL_LOG_URL)'
		url = settings.MINIMAL_LOG_URL
	if check_status:
		assert status in ('good', 'info', 'warn', 'error',), 'not a valid status (provide check_status=False to overrule)'
	data = dict(
		key=key,
		message=message,
		status=status,
	)
	res = post(url=url, data=data)
	if res.status_code != 200:
		raise MinimalLogError('Got {0:}: {1:} when sending log to {2:s} (error was "{3:}")'
			.format(res.status_code, res.content, url, message))






from django.conf import settings
from django.contrib import messages
from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage
from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django.shortcuts import render, redirect
from django.utils.timezone import now
from django.views.decorators.csrf import csrf_exempt
from ipware.ip import get_ip
from .models import MinimalLogEntry, MinimalLogKey


@csrf_exempt
def add_log_entry(request):
	if request.method == 'POST':
		note = ''
		if not request.POST:
			return HttpResponse('POST data did not arrive; was there a redirect?', status=400)
		if not 'status' in request.POST:
			status = request.POST['status']
			note += ' used status "info" since status was not provided;'
		elif not len(request.POST['status']):
			status = request.POST['status']
			note += ' used status "info" since status was empty;'
		elif request.POST['status'] in all_statuses:
			status = request.POST['status']
		else:
			return HttpResponse('status "{0:s}" is not valid (choose one of {1:s})'.format(request.POST['status'],
				', '.join(s for s in all_statuses)), status=400)
		if 'message' in request.POST and len(request.POST['message']) >= 3:
			message = request.POST['message']
		elif 'description' in request.POST and len(request.POST['description']) >= 3:
			message = request.POST['description']
			note += ' note that "description" has been replaced by "message";'
		else:
			return HttpResponse('POST data did not contain a message or message too short', status=400)
		if not 'key' in request.POST:
			return HttpResponse('POST data did not contain a key', status=401)
		try:
			key = MinimalLogKey.objects.get(value=request.POST['key'])
		except MinimalLogKey.DoesNotExist:
			return HttpResponse('the key "{0:s}" is not valid'.format(request.POST['key']), status=401)
		if not key.active:
			return HttpResponse('the key "{0:s}" has been revoked'.format(request.POST['key']), status=401)
		entry = MinimalLogEntry(description=message, status=status, from_ip=get_ip(request), key=key)
		entry.save()
		extra_params = set(request.POST.keys()) - {'status', 'description', 'message', 'key'}
		if extra_params:
			note += ' you provided extra param(s) {0:s} which are redundant'.format(', '.join(s for s in extra_params))
		return HttpResponse('{0:s} added{1:s}'.format(entry.status,
			('; {0:s}\n'.format(note.strip()) if note else '\n')))
	keys = []
	if request.user.is_staff:
		keys = MinimalLogKey.objects.filter(active=True)
	return render(request, 'minimal_log/add_entry.html', {
		'MINIMAL_LOG_TEMPLATE': settings.MINIMAL_LOG_TEMPLATE,
		'statuses': MinimalLogEntry.STATUS_OPTIONS,
		'keys': keys,
	})


def permission_denied(request, action):
	return render(request, 'minimal_log/permission_denied.html', {
		'MINIMAL_LOG_TEMPLATE': settings.MINIMAL_LOG_TEMPLATE,
		'LOGIN_URL': settings.LOGIN_URL,
		'action': action,
	})


def list_log(request):
	"""
		GET parameters (all optional):

		:param page [int]: page number
		:param show [str,str,*]: the types of messages to show
		:param resolved: show resolved messages [no argument]
	"""
	if not (request.user.has_perm('minimal_log_host.change_logentry')):
		return permission_denied(request, 'view')
	entries_list, unresolved_count, show_statuses, show_sources, show_resolved = filtered_entries(
		statuses=request.GET.get('show', ''),
		sources=request.GET.get('from', ''),
		resolved='resolved' in request.GET,
	)
	paginator = Paginator(entries_list, 20)
	page = request.GET.get('page')
	try:
		entries = paginator.page(page)
	except PageNotAnInteger:
		entries = paginator.page(1)
	except EmptyPage:
		entries = paginator.page(paginator.num_pages)
	return render(request, 'minimal_log/list.html', {
		'MINIMAL_LOG_TEMPLATE': settings.MINIMAL_LOG_TEMPLATE,
		'entries': entries,
		'show_sources': ','.join(str(src) for src in sorted(show_sources)),
		'show_statuses': ','.join(show_statuses),
		'show_resolved': show_resolved,
		'pagenr': entries.number,
		'unresolved_count': unresolved_count,
	})


def resolve_log_entry(request):
	if not request.method == 'POST':
		messages.error(request, 'Log entry was not resolved (the request did not have POST data).')
		return redirect(reverse('minimal_log_list'), permanent=False)
	if not (request.user.has_perm('minimal_log_host.change_logentry')):
		return permission_denied(request, 'resolve')
	try:
		entry = MinimalLogEntry.objects.get(pk=int(request.POST['entry']))
	except ValueError:
		return HttpResponse('key "{0:s}" is not an integer'.format(request.POST['entry']))
	except KeyError:
		return HttpResponse('no entry provided')
	except MinimalLogEntry.DoesNotExist:
		return HttpResponse('entry with key "{0:s}" not found'.format(request.POST['entry']))
	action = request.POST.get('action', None)
	if action not in ('resolve', 'unresolve',):
		return HttpResponse('action should be "[un]resolve", not "{0:s}"'.format(action))
	if action == 'resolve':
		entry.resolved = now()
		entry.solver = request.user
		messages.success(request, 'Log entry #{0:d} was resolved.'.format(entry.pk))
	else:
		entry.resolved = None
		entry.solver = None
		messages.success(request, 'Log entry #{0:d} was marked as not resolved.'.format(entry.pk))
	entry.save()
	if request.POST.get('next', None):
		return redirect(to=request.POST['next'], permanent=False)
	return redirect(to=reverse('minimal_log_list'), permanent=False)


def resolve_all_log_entries(request):
	if not request.method == 'POST':
		messages.error(request, 'Log entries were not resolved (the request did not have POST data).')
		return redirect(reverse('minimal_log_list'), permanent=False)
	if not (request.user.has_perm('minimal_log_host.change_logentry')):
		return permission_denied(request, 'resolve')
	entries_list, unresolved_count, show_statuses, show_sources, show_resolved = filtered_entries(
		statuses=request.GET.get('show', ''),
		sources=request.GET.get('from', ''),
		resolved=False,
		sorted=False,
	)
	if not 'confirm' in request.POST and unresolved_count > 3:
		return render(request, 'minimal_log/confirm_resolve.html', {
			'MINIMAL_LOG_TEMPLATE': settings.MINIMAL_LOG_TEMPLATE,
			'count': unresolved_count,
			'next': request.POST.get('next', None),
		})
	entries_list.update(resolved=now(), solver=request.user)
	messages.success(request, '{0:d} log entries were resolved.'.format(unresolved_count))
	if request.POST.get('next', None):
		return redirect(to=request.POST['next'], permanent=False)
	return redirect(to=reverse('minimal_log_list'), permanent=False)


all_statuses = {item[0] for item in MinimalLogEntry.STATUS_OPTIONS}


def filtered_entries(statuses='', resolved=False, sources='', sorted=True):
	show_statuses = [item for item in statuses.split(',') if item in all_statuses]
	entries_list = MinimalLogEntry.objects.all()
	if show_statuses:
		entries_list = entries_list.filter(status__in=show_statuses)
	try:
		show_sources = set(int(source) for source in sources.split(','))
	except ValueError:
		show_sources = set()
	if show_sources:
		entries_list = entries_list.filter(key__pk__in=show_sources)
	unresolved_count = entries_list.filter(resolved__isnull=True).count()
	if not resolved:
		entries_list = entries_list.filter(resolved__isnull=True)
	if sorted:
		entries_list = entries_list.order_by('-added')
	return entries_list, unresolved_count, show_statuses, show_sources, resolved





# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-02-01 23:51
from __future__ import unicode_literals

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import minimal_log_host.utils


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='MinimalLogEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(help_text=b'Describe the problem, change, situation or event.')),
                ('status', models.CharField(choices=[(b'good', b'success'), (b'info', b'info'), (b'warn', b'warning'), (b'error', b'error')], help_text=b'What kind of situation is this?', max_length=8)),
                ('added', models.DateTimeField(auto_now_add=True)),
                ('from_ip', models.CharField(max_length=16, null=True)),
                ('resolved', models.DateTimeField(blank=True, default=None, null=True)),
            ],
            options={
                'verbose_name': 'entry',
                'verbose_name_plural': 'entries',
            },
        ),
        migrations.CreateModel(
            name='MinimalLogKey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(help_text=b"Describe the service uses this key. It's best to generate a new key for each service so that you can easily revoke individual ones.", max_length=128, validators=[django.core.validators.MinLengthValidator(3)])),
                ('value', models.CharField(default=minimal_log_host.utils.generate_key, help_text=b'Secret key used as authentication token (keep it secret!).', max_length=64, unique=True)),
                ('active', models.BooleanField(default=True, help_text=b'Turn this off to disable the key.')),
                ('added', models.DateTimeField(auto_now_add=True)),
                ('adder', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'key',
            },
        ),
        migrations.AddField(
            model_name='minimallogentry',
            name='key',
            field=models.ForeignKey(help_text=b'Key used to authenticate this log entry.', on_delete=django.db.models.deletion.CASCADE, to='minimal_log_host.MinimalLogKey'),
        ),
        migrations.AddField(
            model_name='minimallogentry',
            name='solver',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]




from django.contrib import admin
from django.forms import ModelForm
from .models import MinimalLogKey, MinimalLogEntry


class LogKeyAdminForm(ModelForm):
	"""
	Voodoo to make sure 'adder' is always the user who added or changed the object.
	"""
	def save(self, commit=True, **kwargs):
		instance = super(LogKeyAdminForm, self).save(commit=False, **kwargs)
		instance.adder = self.current_user
		instance.save()
		return instance


class LogKeyAdmin(admin.ModelAdmin):
	readonly_fields = ('added', 'adder', 'last_used',)
	list_display = ('description', 'key_end', 'added', 'adder', 'active', 'last_used',)
	list_display_links = ('description', 'key_end')

	form = LogKeyAdminForm

	def get_form(self, request, obj=None, **kwargs):
		"""
		More voodoo to set 'adder'.
		"""
		form = super(LogKeyAdmin, self).get_form(request, obj, **kwargs)
		form.current_user = request.user
		return form


class LogEntryAdmin(admin.ModelAdmin):
	readonly_fields = ('added', 'from_ip',)
	list_display = ('status', 'description', 'added', 'resolved',)


admin.site.register(MinimalLogKey, LogKeyAdmin)
admin.site.register(MinimalLogEntry, LogEntryAdmin)






from sys import stderr
from logging import Handler
from .python_log import send_log, MinimalLogError, try_get_domain


class MinimalLogHandler(Handler):
	"""
	Sends log messages to minimal_log interface.
	"""

	level_map = {
		'CRITICAL': 'error',
		'ERROR': 'error',
		'WARNING': 'warn',
		'INFO': 'info',
		'DEBUG': 'info',
		'NOTSET': 'error',
	}

	def emit(self, record):
		message = '{0:s} {1:s}: {2:s}'.format(try_get_domain(), record.request.path, record.getMessage())
		status = self.level_map[record.levelname]
		try:
			send_log(message=message, status=status)
		except MinimalLogError as err:
			stderr.write('!! LOGGING ERROR: {0:s}'.format(str(err)))






from django.conf import settings
from django.core.validators import MinLengthValidator
from django.db import models
from django.utils.timesince import timesince
from django.utils.timezone import now
from .utils import generate_key


class MinimalLogKey(models.Model):

	description = models.CharField(max_length=128,  help_text='Describe the service uses this key. It\'s best to generate a new key ' +
		'for each service so that you can easily revoke individual ones.', validators=[MinLengthValidator(3)])
	value = models.CharField(max_length=64, default=generate_key, unique=True,
		help_text='Secret key used as authentication token (keep it secret!).')
	active = models.BooleanField(default=True, help_text='Turn this off to disable the key.')
	added = models.DateTimeField(auto_now_add=True)
	adder = models.ForeignKey(settings.AUTH_USER_MODEL)

	def __str__(self):
		return (self.description or '[no description]') + ' ...{0:s}'.format(self.key_end())

	def __unicode__(self):
		return self.__str__()

	def key_end(self):
		return self.value[-4:]

	def last_used(self):
		entries = MinimalLogEntry.objects.filter(key=self).order_by('-added')
		if entries:
			return timesince(entries[0].added) + ' ago'
		return None

	class Meta:
		verbose_name = 'key'


class MinimalLogEntry(models.Model):
	STATUS_OPTIONS = (('good', 'success'), ('info', 'info'), ('warn', 'warning'), ('error', 'error'))

	description = models.TextField(help_text='Describe the problem, change, situation or event.')
	status = models.CharField(max_length=8, choices=STATUS_OPTIONS, help_text='What kind of situation is this?')
	added = models.DateTimeField(auto_now_add=True)
	from_ip = models.CharField(max_length=16, null=True)
	key = models.ForeignKey(MinimalLogKey, help_text='Key used to authenticate this log entry.')
	resolved = models.DateTimeField(default=None, null=True, blank=True)
	solver = models.ForeignKey(settings.AUTH_USER_MODEL, default=None, blank=True, null=True)

	def __str__(self):
		return '{0:s} {1:s}'.format((self.get_status_display() or '').upper(),
			self.description[:60] + ('...' if len(self.description) > 60 else ''))

	def __unicode__(self):
		return self.__str__()

	def key_end(self):
		return self.key.key_end()

	def resolve(self, request):
		if self.resolved is not None:
			raise AssertionError('log entry %s is already resolved' % self.pk)
		self.resolved = now()
		self.solver = request.user
		self.save()

	def unresolve(self, request):
		if self.resolved is None:
			raise AssertionError('log entry %s is not resolved' % self.pk)
		assert request.user.is_authenticated()
		self.resolved = None
		self.solver = None
		self.save()

	class Meta:
		verbose_name = 'entry'
		verbose_name_plural = 'entries'






from time import time
from django.conf import settings
from django.core.signing import Signer, BadSignature
from django.http import HttpResponse
from django.views.decorators.http import require_POST
from ipware.ip import get_ip
from .python_log import send_log, try_get_domain, MinimalLogError


signer = Signer()
IP_LOG_COUNTER = {}


@require_POST
def receive_log(request, key_lifetime=5*60, per_ip_limit=7, silence_in_debug_mode=True):
	params = {'request_time', 'message', 'csrfmiddlewaretoken'}
	for key in request.POST.keys():
		if key not in params:
			return HttpResponse('unknown parameter "{0:s}"'.format(key), status=400)
	for param in params:
		if param not in request.POST:
			return HttpResponse('parameter "{0:s}" is required'.format(param), status=400)
	try:
		request_time = int(signer.unsign(request.POST['request_time']))
	except BadSignature:
		return HttpResponse('the time signature is not valid'.format(request.POST['key']), status=401)
	if request_time > time() + key_lifetime:
		return HttpResponse('this logging key has expired', status=401)
	ip = get_ip(request)
	message = 'javascript at {0:s} from {2:s}: \n{1:s}'.format(
		try_get_domain(), request.POST['message'][:2048], ip)
	ip_time_hash = '{0:s} {1:d}'.format(ip, int(time() / (60*60*12)))
	if ip_time_hash not in IP_LOG_COUNTER:
		IP_LOG_COUNTER[ip_time_hash] = 1
	IP_LOG_COUNTER[ip_time_hash] += 1
	if IP_LOG_COUNTER[ip_time_hash] > per_ip_limit:
		return HttpResponse('silently ignored due to flooding', status=200)
	if settings.DEBUG and silence_in_debug_mode:
		return HttpResponse('error report discarded because of debug mode', status=200)
	try:
		send_log(message, status='error')
	except MinimalLogError:
		return HttpResponse('there was a server problem while storing the report', status=400)
	else:
		return HttpResponse('ok', status=200)





# -*- coding: utf-8 -*-

"""
	Adapted from https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/
"""

from setuptools import setup


setup(
	name='django_minimal_log',
	description='Simple logging functionality for Django',
	long_description='',
	url='https://github.com/mverleg/django_minimal_log',
	author='Mark V',
	maintainer='(the author)',
	author_email='mdilligaf@gmail.com',
	license='LICENSE.txt',
	keywords=['django', 'logging',],
	version='0.5',
	packages=[
		'minimal_log_host',
		'minimal_logger',
		'minimal_log_host.migrations',
	],
	include_package_data=True,
	zip_safe=False,
	requires=[
		'django',
		'requests',
		'django_ipware',
	],
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Intended Audience :: Developers',
		'Natural Language :: English',
		'License :: OSI Approved :: BSD License',
		'Operating System :: OS Independent',
		'Topic :: System :: Logging',
		'Topic :: System :: Monitoring',
		'Programming Language :: Python',
		'Programming Language :: Python :: 2',
		'Programming Language :: Python :: 2.6',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: Implementation :: PyPy',
		'Topic :: Software Development :: Libraries :: Python Modules',
	],
)






from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.conf.urls import url
from .views import add_log_entry, list_log, resolve_log_entry, resolve_all_log_entries


urlpatterns = [
   url(r'^add/$', add_log_entry, name='minimal_log_add_entry'),
   url(r'^list/$', list_log, name='minimal_log_list'),
   url(r'^resolve/$', resolve_log_entry, name='minimal_log_resolve'),
   url(r'^resolve/all/$', resolve_all_log_entries, name='minimal_log_resolve_all'),
   url(r'^$', lambda request: HttpResponseRedirect(reverse('minimal_log_list'))),
]






from string import ascii_uppercase, digits
from random import SystemRandom


def generate_key(length=32):
	"""
	Generate a random alphanumeric key with '-' and '_' .

	Use length instead of special characters, to prevent possible problems with character encoding from external loggers.
	"""
	return ''.join(SystemRandom().choice(ascii_uppercase + digits + '0123456789--__') for _ in range(32))


